/*
    Toric Lens Converter
    Copyright © 2019 Harry Whitfield

    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Toric Lens Converter - version 2.0
    23 May, 2019
    Copyright © 2019 Harry Whitfield
    mailto:g6auc@arrl.net
*/

/*
Toric Lenses

There is no difference in these forms of notation and it is easy to convert between them:

Add the sphere and cylinder numbers together to produce the converted sphere
Invert the sign of cylinder value
Add 90° to axis value, and if the new axis value exceeds 180°, subtract 180° from the result

For example, a lens with a vertical power of -3.75 and a horizontal power of -2.25 could be specified as either -2.25 -1.50 x 180 or -3.75 +1.50 x 090.
*/

/*jslint this */

/*property
    PI, atan2, axs, convert, cos, cyl, prototype, sin, sph, sqrt, toFixed,
    toString
*/

"use strict";

var print;

function Lens(sph, cyl, axs) {	// constructor
	this.sph = sph;
	this.cyl = cyl || 0;
	this.axs = axs || 0;
}

Lens.prototype.convert = function () {	// convert method
	var sph = this.sph + this.cyl;
	var cyl = -this.cyl;
	var axs = this.axs + 90;
	if (axs >= 180) {
		axs -= 180;
	}
	return new Lens(sph, cyl, axs);
};

Lens.prototype.toString = function () {	// toString method
	return (
		this.sph >= 0
		? "+"
		: ""
	) + this.sph.toFixed(2) + (
		this.cyl !== 0
		? " " + (
			this.cyl >= 0
			? "+"
			: ""
		) + this.cyl.toFixed(2) + " x " + this.axs.toFixed(2)
		: ""
	);
};

function rad(x) {
	return x * Math.PI / 180;
}

function deg(x) {
	return 180 * x / Math.PI;
}

function addLenses(lens1, lens2) {
	var sphere = 0;
	var tmp;
	var fc;
	var f1;
	var f2;
	var alpha;
	var theta;
	var phi;

	sphere += lens1.sph;
	lens1.sph = 0;

	sphere += lens2.sph;
	lens2.sph = 0;

	if (lens1.cyl < 0) {				// convert to + cyl format
		lens1 = lens1.convert();
		sphere += lens1.sph;
		lens1.sph = 0;
	}

	if (lens2.cyl < 0) {				// convert to + cyl format
		lens2 = lens2.convert();
		sphere += lens2.sph;
		lens2.sph = 0;
	}

	if (lens2.axs < lens1.axs) {		// swap lens1 and lens2
		tmp = lens1;
		lens1 = lens2;
		lens2 = tmp;
	}									// lens1.axs now < lens2.axs

	f1 = lens1.cyl;
	f2 = lens2.cyl;

	print("---- ----");

	print("f1: " + f1);
	print("f2: " + f2);

	alpha = lens2.axs - lens1.axs;
	print("alpha: " + alpha);

	fc = Math.sqrt(f1 * f1 + f2 * f2 + 2 * f1 * f2 * Math.cos(rad(2 * alpha)));
	print("fc: " + fc);

	tmp = rad(2 * alpha);
	tmp = Math.atan2(f2 * Math.sin(tmp), f1 + f2 * Math.cos(tmp));		// in -π to π
	theta = deg(0.5 * tmp);
	print("theta: " + theta);											// in -90 to +90

	phi = lens1.axs + theta;
	print("phi: " + phi);

	if (phi < 0) {
		phi += 180;
	}
	if (phi >= 180) {
		phi -= 180;
	}
	sphere += 0.5 * (f1 + f2 - fc);
	return new Lens(sphere, fc, phi);
}

function bicylinder(fc1, a1, fc2, a2) {
	var lens1 = new Lens(0, fc1, a1);
	var lens2 = new Lens(0, fc2, a2);

	return addLenses(lens1, lens2);
}

/*
// Test Code
var lens1 = new Lens(-3.75, 1.50, 90);
var lens2 = lens1.convert();
eprint(lens1 + " === " + lens2);

var lens3 = lens2.convert();
eprint(lens2 + " === " + lens3);

var lens4 = lens1.convert().convert();
eprint(lens1 + " === " + lens4);

var lens5 = new Lens(-3.75);
eprint(lens5);

var lens6 = lens5.convert();
eprint(lens6);

-3.75  1.50  x  90 === -2.25  -1.50  x  180
-2.25  -1.50  x  180 === -3.75  1.50  x  90
-3.75  1.50  x  90 === -3.75  1.50  x  90
-3.75
-3.75

eprint("result: " + bicylinder(2.0, 30, 3.0, 60).toString());	// +0.32 +4.36 x 48.29
eprint("result: " + bicylinder(-2.0, 30, 3.0, 60).toString());	// -0.82 +2.65 x 80.45
eprint("result: " + bicylinder(2.0, 30, 2.0, 120).toString());	// +2.00
*/